"""OpenTherm Protocol message handler"""
import asyncio
import logging
import re
import struct

import pyotgw.messages as m
import pyotgw.vars as v

_LOGGER = logging.getLogger(__name__)


class MessageProcessor:
    """
    Process protocol messages and submit status updates.
    """

    def __init__(
        self,
        command_processor,
        status_manager,
    ):
        """Initialise the protocol object."""
        self._msgq = asyncio.Queue()
        self._task = asyncio.get_running_loop().create_task(self._process_msgs())
        self.command_processor = command_processor
        self.status_manager = status_manager

    async def cleanup(self):
        """Empty the message queue and clean up running task."""
        self.connection_lost()
        if self._task:
            self._task.cancel()
            try:
                await self._task
            except asyncio.CancelledError:
                self._task = None

    def connection_lost(self):
        """
        Gets called when the connection to the gateway is lost.
        Tear down and clean up the object.
        """
        while not self._msgq.empty():
            self._msgq.get_nowait()

    def submit_matched_message(self, match):
        """Add a matched message to the processing queue."""
        src, mtype, mid, msb, lsb = self._dissect_msg(match)
        if lsb is not None:
            self._msgq.put_nowait((src, mtype, mid, msb, lsb))
            _LOGGER.debug(
                "Added line to message queue. Queue size: %d",
                self._msgq.qsize(),
            )

    def _dissect_msg(self, match):
        """
        Split messages into bytes and return a tuple of bytes.
        """
        recvfrom = match.group(1)
        frame = bytes.fromhex(match.group(2))
        if recvfrom == "E":
            _LOGGER.info(
                "The OpenTherm Gateway received an erroneous message."
                " This is not a bug in pyotgw. Ignoring: %s",
                frame.hex().upper(),
            )
            return None, None, None, None, None
        msgtype = self._get_msgtype(frame[0])
        if msgtype in (v.READ_ACK, v.WRITE_ACK, v.READ_DATA, v.WRITE_DATA):
            # Some info is best read from the READ/WRITE_DATA messages
            # as the boiler may not support the data ID.
            # Slice syntax is used to prevent implicit cast to int.
            data_id = frame[1:2]
            data_msb = frame[2:3]
            data_lsb = frame[3:4]
            return recvfrom, msgtype, data_id, data_msb, data_lsb
        return None, None, None, None, None

    @staticmethod
    def _get_msgtype(byte):
        """
        Return the message type of Opentherm messages according to
        byte.
        """
        return (byte >> 4) & 0x7

    async def _process_msgs(self):
        """
        Get messages from the queue and pass them to _process_msg().
        Make sure we process one message at a time to keep them in sequence.
        """
        while True:
            args = await self._msgq.get()
            _LOGGER.debug(
                "Processing: %s %02x %s %s %s",
                args[0],
                args[1],
                *[args[i].hex().upper() for i in range(2, 5)],
            )
            await self._process_msg(args)

    async def _process_msg(self, message):
        """
        Process message and update status variables where necessary.
        Add status to queue if it was changed in the process.
        """
        (
            src,
            mtype,
            msgid,
            msb,  # pylint: disable=possibly-unused-variable
            lsb,  # pylint: disable=possibly-unused-variable
        ) = message
        if msgid not in m.REGISTRY:
            return

        if src in "TA":
            part = v.THERMOSTAT
        else:  # src in "BR"
            part = v.BOILER
        update = {}

        for action in m.REGISTRY[msgid][m.MSG_TYPE[mtype]]:
            update.update(await self._get_dict_update_for_action(action, locals()))

        if not update:
            return

        self.status_manager.submit_partial_update(part, update)

    async def _get_dict_update_for_action(self, action, env):
        """Return a partial dict update for message"""
        func = getattr(self, action[m.FUNC])
        loc = locals()
        loc.update(env)
        args = (loc[arg] for arg in action[m.ARGS])
        if asyncio.iscoroutinefunction(func):
            ret = await func(*args)
        else:
            ret = func(*args)
        ret = ret if isinstance(ret, list) else [ret]
        update = {}
        for var, val in zip(action[m.RETURNS], ret):
            if var is False:
                return {}
            if var is None:
                continue
            update.update({var: val})
        return update

    async def _quirk_trovrd(self, part, src, msb, lsb):
        """Handle MSG_TROVRD with iSense quirk"""
        update = {}
        ovrd_value = self._get_f8_8(msb, lsb)
        if ovrd_value > 0:
            # iSense quirk: the gateway keeps sending override value
            # even if the thermostat has cancelled the override.
            if (
                self.status_manager.status[v.OTGW].get(v.OTGW_THRM_DETECT) == "I"
                and src == "A"
            ):
                ovrd = await self.command_processor.issue_cmd(
                    v.OTGW_CMD_REPORT, v.OTGW_REPORT_SETPOINT_OVRD
                )
                match = re.match(r"^O=(N|[CT]([0-9]+.[0-9]+))$", ovrd, re.IGNORECASE)
                if not match:
                    return
                if match.group(1) in "Nn":
                    self.status_manager.delete_value(part, v.DATA_ROOM_SETPOINT_OVRD)
                    return
                update[v.DATA_ROOM_SETPOINT_OVRD] = float(match.group(2))
            else:
                update[v.DATA_ROOM_SETPOINT_OVRD] = ovrd_value
            self.status_manager.submit_partial_update(part, update)
        else:
            self.status_manager.delete_value(part, v.DATA_ROOM_SETPOINT_OVRD)

    async def _quirk_trset_s2m(self, part, msb, lsb):
        """Handle MSG_TRSET with gateway quirk"""
        # Ignore s2m messages on thermostat side as they are ALWAYS WriteAcks
        # but may contain invalid data.
        if part == v.THERMOSTAT:
            return

        self.status_manager.submit_partial_update(
            part, {v.DATA_ROOM_SETPOINT: self._get_f8_8(msb, lsb)}
        )

    @staticmethod
    def _get_flag8(byte):
        """
        Split a byte into a list of 8 bits (1/0).
        """
        ret = [0, 0, 0, 0, 0, 0, 0, 0]
        byte = byte[0]
        for i in range(0, 8):
            ret[i] = byte & 1
            byte = byte >> 1
        return ret

    @staticmethod
    def _get_u8(byte):
        """
        Convert a byte into an unsigned int.
        """
        return struct.unpack(">B", byte)[0]

    @staticmethod
    def _get_s8(byte):
        """
        Convert a byte into a signed int.
        """
        return struct.unpack(">b", byte)[0]

    def _get_f8_8(self, msb, lsb):
        """
        Convert 2 bytes into an OpenTherm f8_8 (float) value.
        """
        return float(self._get_s16(msb, lsb) / 256)

    def _get_u16(self, msb, lsb):
        """
        Convert 2 bytes into an unsigned int.
        """
        buf = struct.pack(">BB", self._get_u8(msb), self._get_u8(lsb))
        return int(struct.unpack(">H", buf)[0])

    def _get_s16(self, msb, lsb):
        """
        Convert 2 bytes into a signed int.
        """
        buf = struct.pack(">bB", self._get_s8(msb), self._get_u8(lsb))
        return int(struct.unpack(">h", buf)[0])



"""Data related to message processing"""

from pyotgw import vars as v

_GET_FLAG8 = "_get_flag8"
_GET_FLOAT = "_get_f8_8"
_GET_S16 = "_get_s16"
_GET_S8 = "_get_s8"
_GET_U16 = "_get_u16"
_GET_U8 = "_get_u8"
_LSB = "lsb"
_MSB = "msb"

ARGS = "args"
FUNC = "function"
M2S = "m2s"
RETURNS = "returns"
S2M = "s2m"

MSG_TYPE = {
    v.READ_DATA: M2S,
    v.WRITE_DATA: M2S,
    v.READ_ACK: S2M,
    v.WRITE_ACK: S2M,
}

REGISTRY = {
    # MSG_ID = {
    #     msg_type: [
    #         {
    #             FUNC: "func_name",
    #             ARGS: (f_args,),
    #             RETURNS: (val_1, val_2, ..., val_n),
    #         },
    #     ],
    # }
    v.MSG_STATUS: {
        M2S: [
            {
                FUNC: _GET_FLAG8,
                ARGS: (_MSB,),
                RETURNS: (
                    v.DATA_MASTER_CH_ENABLED,
                    v.DATA_MASTER_DHW_ENABLED,
                    v.DATA_MASTER_COOLING_ENABLED,
                    v.DATA_MASTER_OTC_ENABLED,
                    v.DATA_MASTER_CH2_ENABLED,
                ),
            },
        ],
        S2M: [
            {
                FUNC: _GET_FLAG8,
                ARGS: (_LSB,),
                RETURNS: (
                    v.DATA_SLAVE_FAULT_IND,
                    v.DATA_SLAVE_CH_ACTIVE,
                    v.DATA_SLAVE_DHW_ACTIVE,
                    v.DATA_SLAVE_FLAME_ON,
                    v.DATA_SLAVE_COOLING_ACTIVE,
                    v.DATA_SLAVE_CH2_ACTIVE,
                    v.DATA_SLAVE_DIAG_IND,
                ),
            },
        ],
    },
    v.MSG_TSET: {
        M2S: [],
        S2M: [
            {
                FUNC: _GET_FLOAT,
                ARGS: (
                    _MSB,
                    _LSB,
                ),
                RETURNS: (v.DATA_CONTROL_SETPOINT,),
            },
        ],
    },
    v.MSG_MCONFIG: {
        M2S: [{FUNC: _GET_U8, ARGS: (_LSB,), RETURNS: (v.DATA_MASTER_MEMBERID,)}],
        S2M: [],
    },
    v.MSG_SCONFIG: {
        M2S: [],
        S2M: [
            {
                FUNC: _GET_FLAG8,
                ARGS: (_MSB,),
                RETURNS: (
                    v.DATA_SLAVE_DHW_PRESENT,
                    v.DATA_SLAVE_CONTROL_TYPE,
                    v.DATA_SLAVE_COOLING_SUPPORTED,
                    v.DATA_SLAVE_DHW_CONFIG,
                    v.DATA_SLAVE_MASTER_LOW_OFF_PUMP,
                    v.DATA_SLAVE_CH2_PRESENT,
                ),
            },
            {FUNC: _GET_U8, ARGS: (_LSB,), RETURNS: (v.DATA_SLAVE_MEMBERID,)},
        ],
    },
    v.MSG_COMMAND: {M2S: [], S2M: []},
    v.MSG_ASFFLAGS: {
        M2S: [],
        S2M: [
            {
                FUNC: _GET_FLAG8,
                ARGS: (_MSB,),
                RETURNS: (
                    v.DATA_SLAVE_SERVICE_REQ,
                    v.DATA_SLAVE_REMOTE_RESET,
                    v.DATA_SLAVE_LOW_WATER_PRESS,
                    v.DATA_SLAVE_GAS_FAULT,
                    v.DATA_SLAVE_AIR_PRESS_FAULT,
                    v.DATA_SLAVE_WATER_OVERTEMP,
                ),
            },
            {FUNC: _GET_U8, ARGS: (_LSB,), RETURNS: (v.DATA_SLAVE_OEM_FAULT,)},
        ],
    },
    v.MSG_RBPFLAGS: {
        M2S: [],
        S2M: [
            {
                FUNC: _GET_FLAG8,
                ARGS: (_MSB,),
                RETURNS: (
                    v.DATA_REMOTE_TRANSFER_DHW,
                    v.DATA_REMOTE_TRANSFER_MAX_CH,
                ),
            },
            {
                FUNC: _GET_FLAG8,
                ARGS: (_LSB,),
                RETURNS: (
                    v.DATA_REMOTE_RW_DHW,
                    v.DATA_REMOTE_RW_MAX_CH,
                ),
            },
        ],
    },
    v.MSG_COOLING: {
        M2S: [],
        S2M: [
            {
                FUNC: _GET_FLOAT,
                ARGS: (
                    _MSB,
                    _LSB,
                ),
                RETURNS: (v.DATA_COOLING_CONTROL,),
            },
        ],
    },
    v.MSG_TSETC2: {
        M2S: [],
        S2M: [
            {
                FUNC: _GET_FLOAT,
                ARGS: (
                    _MSB,
                    _LSB,
                ),
                RETURNS: (v.DATA_CONTROL_SETPOINT_2,),
            },
        ],
    },
    v.MSG_TROVRD: {
        M2S: [],
        S2M: [
            {
                FUNC: "_quirk_trovrd",
                ARGS: (
                    "part",
                    "src",
                    _MSB,
                    _LSB,
                ),
                RETURNS: (False,),
            },
        ],
    },
    v.MSG_TSP: {M2S: [], S2M: []},
    v.MSG_TSPIDX: {M2S: [], S2M: []},
    v.MSG_FHBSIZE: {M2S: [], S2M: []},
    v.MSG_FHBIDX: {M2S: [], S2M: []},
    v.MSG_MAXRMOD: {
        M2S: [],
        S2M: [
            {
                FUNC: _GET_FLOAT,
                ARGS: (
                    _MSB,
                    _LSB,
                ),
                RETURNS: (v.DATA_SLAVE_MAX_RELATIVE_MOD,),
            },
        ],
    },
    v.MSG_MAXCAPMINMOD: {
        M2S: [],
        S2M: [
            {FUNC: _GET_U8, ARGS: (_MSB,), RETURNS: (v.DATA_SLAVE_MAX_CAPACITY,)},
            {FUNC: _GET_U8, ARGS: (_LSB,), RETURNS: (v.DATA_SLAVE_MIN_MOD_LEVEL,)},
        ],
    },
    v.MSG_TRSET: {
        M2S: [
            {
                FUNC: _GET_FLOAT,
                ARGS: (
                    _MSB,
                    _LSB,
                ),
                RETURNS: (v.DATA_ROOM_SETPOINT,),
            },
        ],
        S2M: [
            {
                FUNC: "_quirk_trset_s2m",
                ARGS: (
                    "part",
                    _MSB,
                    _LSB,
                ),
                RETURNS: (False,),
            },
        ],
    },
    v.MSG_RELMOD: {
        M2S: [],
        S2M: [
            {
                FUNC: _GET_FLOAT,
                ARGS: (
                    _MSB,
                    _LSB,
                ),
                RETURNS: (v.DATA_REL_MOD_LEVEL,),
            },
        ],
    },
    v.MSG_CHPRESS: {
        M2S: [],
        S2M: [
            {
                FUNC: _GET_FLOAT,
                ARGS: (
                    _MSB,
                    _LSB,
                ),
                RETURNS: (v.DATA_CH_WATER_PRESS,),
            },
        ],
    },
    v.MSG_DHWFLOW: {
        M2S: [],
        S2M: [
            {
                FUNC: _GET_FLOAT,
                ARGS: (
                    _MSB,
                    _LSB,
                ),
                RETURNS: (v.DATA_DHW_FLOW_RATE,),
            },
        ],
    },
    v.MSG_TIME: {M2S: [], S2M: []},
    v.MSG_DATE: {M2S: [], S2M: []},
    v.MSG_YEAR: {M2S: [], S2M: []},
    v.MSG_TRSET2: {
        M2S: [
            {
                FUNC: _GET_FLOAT,
                ARGS: (
                    _MSB,
                    _LSB,
                ),
                RETURNS: (v.DATA_ROOM_SETPOINT_2,),
            },
        ],
        S2M: [],
    },
    v.MSG_TROOM: {
        M2S: [
            {
                FUNC: _GET_FLOAT,
                ARGS: (
                    _MSB,
                    _LSB,
                ),
                RETURNS: (v.DATA_ROOM_TEMP,),
            }
        ],
        S2M: [],
    },
    v.MSG_TBOILER: {
        M2S: [],
        S2M: [
            {
                FUNC: _GET_FLOAT,
                ARGS: (
                    _MSB,
                    _LSB,
                ),
                RETURNS: (v.DATA_CH_WATER_TEMP,),
            },
        ],
    },
    v.MSG_TDHW: {
        M2S: [],
        S2M: [
            {
                FUNC: _GET_FLOAT,
                ARGS: (
                    _MSB,
                    _LSB,
                ),
                RETURNS: (v.DATA_DHW_TEMP,),
            }
        ],
    },
    v.MSG_TOUTSIDE: {
        M2S: [],
        S2M: [
            {
                FUNC: _GET_FLOAT,
                ARGS: (
                    _MSB,
                    _LSB,
                ),
                RETURNS: (v.DATA_OUTSIDE_TEMP,),
            },
        ],
    },
    v.MSG_TRET: {
        M2S: [],
        S2M: [
            {
                FUNC: _GET_FLOAT,
                ARGS: (
                    _MSB,
                    _LSB,
                ),
                RETURNS: (v.DATA_RETURN_WATER_TEMP,),
            },
        ],
    },
    v.MSG_TSTOR: {
        M2S: [],
        S2M: [
            {
                FUNC: _GET_FLOAT,
                ARGS: (
                    _MSB,
                    _LSB,
                ),
                RETURNS: (v.DATA_SOLAR_STORAGE_TEMP,),
            },
        ],
    },
    v.MSG_TCOLL: {
        M2S: [],
        S2M: [
            {
                FUNC: _GET_FLOAT,
                ARGS: (
                    _MSB,
                    _LSB,
                ),
                RETURNS: (v.DATA_SOLAR_COLL_TEMP,),
            },
        ],
    },
    v.MSG_TFLOWCH2: {
        M2S: [],
        S2M: [
            {
                FUNC: _GET_FLOAT,
                ARGS: (
                    _MSB,
                    _LSB,
                ),
                RETURNS: (v.DATA_CH_WATER_TEMP_2,),
            },
        ],
    },
    v.MSG_TDHW2: {
        M2S: [],
        S2M: [
            {
                FUNC: _GET_FLOAT,
                ARGS: (
                    _MSB,
                    _LSB,
                ),
                RETURNS: (v.DATA_DHW_TEMP_2,),
            }
        ],
    },
    v.MSG_TEXHAUST: {
        M2S: [],
        S2M: [
            {
                FUNC: _GET_S16,
                ARGS: (
                    _MSB,
                    _LSB,
                ),
                RETURNS: (v.DATA_EXHAUST_TEMP,),
            }
        ],
    },
    v.MSG_TDHWSETUL: {
        M2S: [],
        S2M: [
            {FUNC: _GET_S8, ARGS: (_MSB,), RETURNS: (v.DATA_SLAVE_DHW_MAX_SETP,)},
            {FUNC: _GET_S8, ARGS: (_LSB,), RETURNS: (v.DATA_SLAVE_DHW_MIN_SETP,)},
        ],
    },
    v.MSG_TCHSETUL: {
        M2S: [],
        S2M: [
            {FUNC: _GET_S8, ARGS: (_MSB,), RETURNS: (v.DATA_SLAVE_CH_MAX_SETP,)},
            {FUNC: _GET_S8, ARGS: (_LSB,), RETURNS: (v.DATA_SLAVE_CH_MIN_SETP,)},
        ],
    },
    v.MSG_OTCCURVEUL: {M2S: [], S2M: []},
    v.MSG_TDHWSET: {
        M2S: [],
        S2M: [
            {
                FUNC: _GET_FLOAT,
                ARGS: (
                    _MSB,
                    _LSB,
                ),
                RETURNS: (v.DATA_DHW_SETPOINT,),
            },
        ],
    },
    v.MSG_MAXTSET: {
        M2S: [],
        S2M: [
            {
                FUNC: _GET_FLOAT,
                ARGS: (
                    _MSB,
                    _LSB,
                ),
                RETURNS: (v.DATA_MAX_CH_SETPOINT,),
            },
        ],
    },
    v.MSG_OTCCURVE: {M2S: [], S2M: []},
    v.MSG_STATUSVH: {
        M2S: [
            {
                FUNC: _GET_FLAG8,
                ARGS: (_MSB,),
                RETURNS: (
                    v.DATA_VH_MASTER_VENT_ENABLED,
                    v.DATA_VH_MASTER_BYPASS_POS,
                    v.DATA_VH_MASTER_BYPASS_MODE,
                    v.DATA_VH_MASTER_FREE_VENT_MODE,
                ),
            },
        ],
        S2M: [
            {
                FUNC: _GET_FLAG8,
                ARGS: (_LSB,),
                RETURNS: (
                    v.DATA_VH_SLAVE_FAULT_INDICATE,
                    v.DATA_VH_SLAVE_VENT_MODE,
                    v.DATA_VH_SLAVE_BYPASS_STATUS,
                    v.DATA_VH_SLAVE_BYPASS_AUTO_STATUS,
                    v.DATA_VH_SLAVE_FREE_VENT_STATUS,
                    None,
                    v.DATA_VH_SLAVE_DIAG_INDICATE,
                ),
            },
        ],
    },
    v.MSG_RELVENTPOS: {
        M2S: [{FUNC: _GET_U8, ARGS: (_MSB,), RETURNS: (v.DATA_VH_CONTROL_SETPOINT,)}],
        S2M: [],
    },
    v.MSG_RELVENT: {
        M2S: [],
        S2M: [{FUNC: _GET_U8, ARGS: (_MSB,), RETURNS: (v.DATA_VH_RELATIVE_VENT,)}],
    },
    v.MSG_ROVRD: {
        M2S: [],
        S2M: [
            {
                FUNC: _GET_FLAG8,
                ARGS: (_LSB,),
                RETURNS: (
                    v.DATA_ROVRD_MAN_PRIO,
                    v.DATA_ROVRD_AUTO_PRIO,
                ),
            },
        ],
    },
    v.MSG_OEMDIAG: {
        M2S: [],
        S2M: [
            {
                FUNC: _GET_U16,
                ARGS: (
                    _MSB,
                    _LSB,
                ),
                RETURNS: (v.DATA_OEM_DIAG,),
            }
        ],
    },
    v.MSG_BURNSTARTS: {
        M2S: [],
        S2M: [
            {
                FUNC: _GET_U16,
                ARGS: (
                    _MSB,
                    _LSB,
                ),
                RETURNS: (v.DATA_TOTAL_BURNER_STARTS,),
            },
        ],
    },
    v.MSG_CHPUMPSTARTS: {
        M2S: [],
        S2M: [
            {
                FUNC: _GET_U16,
                ARGS: (
                    _MSB,
                    _LSB,
                ),
                RETURNS: (v.DATA_CH_PUMP_STARTS,),
            },
        ],
    },
    v.MSG_DHWPUMPSTARTS: {
        M2S: [],
        S2M: [
            {
                FUNC: _GET_U16,
                ARGS: (
                    _MSB,
                    _LSB,
                ),
                RETURNS: (v.DATA_DHW_PUMP_STARTS,),
            },
        ],
    },
    v.MSG_DHWBURNSTARTS: {
        M2S: [],
        S2M: [
            {
                FUNC: _GET_U16,
                ARGS: (
                    _MSB,
                    _LSB,
                ),
                RETURNS: (v.DATA_DHW_BURNER_STARTS,),
            },
        ],
    },
    v.MSG_BURNHRS: {
        M2S: [],
        S2M: [
            {
                FUNC: _GET_U16,
                ARGS: (
                    _MSB,
                    _LSB,
                ),
                RETURNS: (v.DATA_TOTAL_BURNER_HOURS,),
            },
        ],
    },
    v.MSG_CHPUMPHRS: {
        M2S: [],
        S2M: [
            {
                FUNC: _GET_U16,
                ARGS: (
                    _MSB,
                    _LSB,
                ),
                RETURNS: (v.DATA_CH_PUMP_HOURS,),
            }
        ],
    },
    v.MSG_DHWPUMPHRS: {
        M2S: [],
        S2M: [
            {
                FUNC: _GET_U16,
                ARGS: (
                    _MSB,
                    _LSB,
                ),
                RETURNS: (v.DATA_DHW_PUMP_HOURS,),
            },
        ],
    },
    v.MSG_DHWBURNHRS: {
        M2S: [],
        S2M: [
            {
                FUNC: _GET_U16,
                ARGS: (
                    _MSB,
                    _LSB,
                ),
                RETURNS: (v.DATA_DHW_BURNER_HOURS,),
            },
        ],
    },
    v.MSG_OTVERM: {
        M2S: [
            {
                FUNC: _GET_FLOAT,
                ARGS: (
                    _MSB,
                    _LSB,
                ),
                RETURNS: (v.DATA_MASTER_OT_VERSION,),
            },
        ],
        S2M: [],
    },
    v.MSG_OTVERS: {
        M2S: [],
        S2M: [
            {
                FUNC: _GET_FLOAT,
                ARGS: (
                    _MSB,
                    _LSB,
                ),
                RETURNS: (v.DATA_SLAVE_OT_VERSION,),
            },
        ],
    },
    v.MSG_MVER: {
        M2S: [
            {FUNC: _GET_U8, ARGS: (_MSB,), RETURNS: (v.DATA_MASTER_PRODUCT_TYPE,)},
            {FUNC: _GET_U8, ARGS: (_LSB,), RETURNS: (v.DATA_MASTER_PRODUCT_VERSION,)},
        ],
        S2M: [],
    },
    v.MSG_SVER: {
        M2S: [],
        S2M: [
            {FUNC: _GET_U8, ARGS: (_MSB,), RETURNS: (v.DATA_SLAVE_PRODUCT_TYPE,)},
            {FUNC: _GET_U8, ARGS: (_LSB,), RETURNS: (v.DATA_SLAVE_PRODUCT_VERSION,)},
        ],
    },
}



"""All status related code"""

import asyncio
import logging
from copy import deepcopy

from pyotgw import vars as v

_LOGGER = logging.getLogger(__name__)


class StatusManager:
    """Manage status tracking and updates"""

    def __init__(self):
        """Initialise the status manager"""
        self.loop = asyncio.get_event_loop()
        self._updateq = asyncio.Queue()
        self._status = deepcopy(v.DEFAULT_STATUS)
        self._notify = []
        self._update_task = self.loop.create_task(self._process_updates())

    def reset(self):
        """Clear the queue and reset the status dict"""
        while not self._updateq.empty():
            self._updateq.get_nowait()
        self._status = deepcopy(v.DEFAULT_STATUS)

    @property
    def status(self):
        """Return the full status dict"""
        return deepcopy(self._status)

    def delete_value(self, part, key):
        """Delete key from status part."""
        try:
            del self._status[part][key]
        except (AttributeError, KeyError):
            return False
        self._updateq.put_nowait(self.status)
        return True

    def submit_partial_update(self, part, update):
        """
        Submit an update for part of the status dict to the queue.
        Return a boolean indicating success.
        """
        if part not in self.status:
            _LOGGER.error("Invalid status part for update: %s", part)
            return False
        if not isinstance(update, dict):
            _LOGGER.error("Update for %s is not a dict: %s", part, update)
            return False
        self._status[part].update(update)
        self._updateq.put_nowait(self.status)
        return True

    def submit_full_update(self, update):
        """
        Submit an update for multiple parts of the status dict to the
        queue. Return a boolean indicating success.
        """
        for part, values in update.items():
            # First we verify all data
            if part not in self.status:
                _LOGGER.error("Invalid status part for update: %s", part)
                return False
            if not isinstance(values, dict):
                _LOGGER.error("Update for %s is not a dict: %s", part, values)
                return False
        for part, values in update.items():
            # Then we actually update
            self._status[part].update(values)
        self._updateq.put_nowait(self.status)
        return True

    def subscribe(self, callback):
        """
        Subscribe callback for future status updates.
        Return boolean indicating success.
        """
        if callback in self._notify:
            return False
        self._notify.append(callback)
        return True

    def unsubscribe(self, callback):
        """
        Unsubscribe callback from future status updates.
        Return boolean indicating success.
        """
        if callback not in self._notify:
            return False
        self._notify.remove(callback)
        return True

    async def cleanup(self):
        """Clean up task"""
        if self._update_task:
            self._update_task.cancel()
            try:
                await self._update_task
            except asyncio.CancelledError:
                self._update_task = None

    async def _process_updates(self):
        """Process updates from the queue."""
        _LOGGER.debug("Starting reporting routine")
        while True:
            oldstatus = deepcopy(self.status)
            stat = await self._updateq.get()
            if oldstatus != stat and self._notify:
                for coro in self._notify:
                    # Each client gets its own copy of the dict.
                    self.loop.create_task(coro(deepcopy(stat)))



"""Test data for pyotgw tests"""

from types import SimpleNamespace

import pyotgw.vars as v

_report_responses_51 = {
    v.OTGW_REPORT_ABOUT: "A=OpenTherm Gateway 5.1",
    v.OTGW_REPORT_BUILDDATE: "B=17:44 11-02-2021",
    v.OTGW_REPORT_CLOCKMHZ: "C=4 MHz",
    v.OTGW_REPORT_DHW_SETTING: "W=A",
    v.OTGW_REPORT_GPIO_FUNCS: "G=10",
    v.OTGW_REPORT_GPIO_STATES: "I=00",
    v.OTGW_REPORT_LED_FUNCS: "L=FXOMPC",
    v.OTGW_REPORT_GW_MODE: "M=G",
    v.OTGW_REPORT_RST_CAUSE: "Q=C",
    v.OTGW_REPORT_SETBACK_TEMP: "S=16.50",
    v.OTGW_REPORT_SETPOINT_OVRD: "O=T20.50",
    v.OTGW_REPORT_SMART_PWR: "P=Low power",
    v.OTGW_REPORT_TEMP_SENSOR: "D=0",
    v.OTGW_REPORT_THERMOSTAT_DETECT: "R=D",
    v.OTGW_REPORT_TWEAKS: "T=11",
    v.OTGW_REPORT_VREF: "V=3",
}

_report_responses_42 = {
    v.OTGW_REPORT_ABOUT: "A=OpenTherm Gateway 4.2.5",
    v.OTGW_REPORT_BUILDDATE: "B=17:59 20-10-2015",
    v.OTGW_REPORT_CLOCKMHZ: None,
    v.OTGW_REPORT_DHW_SETTING: "W=A",
    v.OTGW_REPORT_GPIO_FUNCS: "G=10",
    v.OTGW_REPORT_GPIO_STATES: "I=00",
    v.OTGW_REPORT_LED_FUNCS: "L=FXOMPC",
    v.OTGW_REPORT_GW_MODE: "M=G",
    v.OTGW_REPORT_RST_CAUSE: "Q=C",
    v.OTGW_REPORT_SETBACK_TEMP: "S=16.50",
    v.OTGW_REPORT_SETPOINT_OVRD: "O=T20.50",
    v.OTGW_REPORT_SMART_PWR: "P=Low power",
    v.OTGW_REPORT_THERMOSTAT_DETECT: "R=D",
    v.OTGW_REPORT_TWEAKS: "T=11",
    v.OTGW_REPORT_VREF: "V=3",
}

_report_expect_51 = {
    v.BOILER: {},
    v.OTGW: {
        v.OTGW_ABOUT: "OpenTherm Gateway 5.1",
        v.OTGW_BUILD: "17:44 11-02-2021",
        v.OTGW_CLOCKMHZ: "4 MHz",
        v.OTGW_DHW_OVRD: "A",
        v.OTGW_MODE: "G",
        v.OTGW_RST_CAUSE: "C",
        v.OTGW_SMART_PWR: "Low power",
        v.OTGW_TEMP_SENSOR: "0",
        v.OTGW_THRM_DETECT: "D",
        v.OTGW_SETP_OVRD_MODE: "T",
        v.OTGW_GPIO_A: 1,
        v.OTGW_GPIO_B: 0,
        v.OTGW_LED_A: "F",
        v.OTGW_LED_B: "X",
        v.OTGW_LED_C: "O",
        v.OTGW_LED_D: "M",
        v.OTGW_LED_E: "P",
        v.OTGW_LED_F: "C",
        v.OTGW_IGNORE_TRANSITIONS: 1,
        v.OTGW_OVRD_HB: 1,
        v.OTGW_SB_TEMP: 16.5,
        v.OTGW_VREF: 3,
    },
    v.THERMOSTAT: {v.DATA_ROOM_SETPOINT_OVRD: 20.5},
}

_report_expect_42 = {
    v.BOILER: {},
    v.OTGW: {
        v.OTGW_ABOUT: "OpenTherm Gateway 4.2.5",
        v.OTGW_BUILD: "17:59 20-10-2015",
        v.OTGW_CLOCKMHZ: None,
        v.OTGW_DHW_OVRD: "A",
        v.OTGW_MODE: "G",
        v.OTGW_RST_CAUSE: "C",
        v.OTGW_SMART_PWR: "Low power",
        v.OTGW_TEMP_SENSOR: None,
        v.OTGW_THRM_DETECT: "D",
        v.OTGW_SETP_OVRD_MODE: "T",
        v.OTGW_GPIO_A: 1,
        v.OTGW_GPIO_B: 0,
        v.OTGW_LED_A: "F",
        v.OTGW_LED_B: "X",
        v.OTGW_LED_C: "O",
        v.OTGW_LED_D: "M",
        v.OTGW_LED_E: "P",
        v.OTGW_LED_F: "C",
        v.OTGW_IGNORE_TRANSITIONS: 1,
        v.OTGW_OVRD_HB: 1,
        v.OTGW_SB_TEMP: 16.5,
        v.OTGW_VREF: 3,
    },
    v.THERMOSTAT: {v.DATA_ROOM_SETPOINT_OVRD: 20.5},
}

pygw_reports = SimpleNamespace(
    expect_42=_report_expect_42,
    expect_51=_report_expect_51,
    report_responses_42=_report_responses_42,
    report_responses_51=_report_responses_51,
)


_status_5 = (
    "10101010/01010101,1.23,01010101/10101010,2.34,3.45,4.56,1/0,5.67,6.78,7.89,8.90,"
    "9.09,0.98,9.87,8.76,7.65,6.54,5.43,2,34/56,78/90,4.32,3.21,10101010/01010101,9,8,"
    "7654,6543,54321,43210,32101,21012,10123,99"
)
_status_expect_5 = {
    v.BOILER: {
        v.DATA_SLAVE_FAULT_IND: 1,
        v.DATA_SLAVE_CH_ACTIVE: 0,
        v.DATA_SLAVE_DHW_ACTIVE: 1,
        v.DATA_SLAVE_FLAME_ON: 0,
        v.DATA_SLAVE_COOLING_ACTIVE: 1,
        v.DATA_SLAVE_CH2_ACTIVE: 0,
        v.DATA_SLAVE_DIAG_IND: 1,
        v.DATA_REMOTE_TRANSFER_DHW: 1,
        v.DATA_REMOTE_TRANSFER_MAX_CH: 0,
        v.DATA_REMOTE_RW_DHW: 0,
        v.DATA_REMOTE_RW_MAX_CH: 1,
        v.DATA_SLAVE_MAX_RELATIVE_MOD: 4.56,
        v.DATA_SLAVE_MAX_CAPACITY: 1,
        v.DATA_SLAVE_MIN_MOD_LEVEL: 0,
        v.DATA_REL_MOD_LEVEL: 6.78,
        v.DATA_CH_WATER_PRESS: 7.89,
        v.DATA_DHW_FLOW_RATE: 8.90,
        v.DATA_CH_WATER_TEMP: 9.87,
        v.DATA_DHW_TEMP: 8.76,
        v.DATA_OUTSIDE_TEMP: 7.65,
        v.DATA_RETURN_WATER_TEMP: 6.54,
        v.DATA_CH_WATER_TEMP_2: 5.43,
        v.DATA_EXHAUST_TEMP: 2,
        v.DATA_SLAVE_DHW_MAX_SETP: 34,
        v.DATA_SLAVE_DHW_MIN_SETP: 56,
        v.DATA_SLAVE_CH_MAX_SETP: 78,
        v.DATA_SLAVE_CH_MIN_SETP: 90,
        v.DATA_DHW_SETPOINT: 4.32,
        v.DATA_MAX_CH_SETPOINT: 3.21,
        v.DATA_VH_SLAVE_FAULT_INDICATE: 1,
        v.DATA_VH_SLAVE_VENT_MODE: 0,
        v.DATA_VH_SLAVE_BYPASS_STATUS: 1,
        v.DATA_VH_SLAVE_BYPASS_AUTO_STATUS: 0,
        v.DATA_VH_SLAVE_FREE_VENT_STATUS: 1,
        v.DATA_VH_SLAVE_DIAG_INDICATE: 1,
        v.DATA_VH_RELATIVE_VENT: 8,
        v.DATA_TOTAL_BURNER_STARTS: 7654,
        v.DATA_CH_PUMP_STARTS: 6543,
        v.DATA_DHW_PUMP_STARTS: 54321,
        v.DATA_DHW_BURNER_STARTS: 43210,
        v.DATA_TOTAL_BURNER_HOURS: 32101,
        v.DATA_CH_PUMP_HOURS: 21012,
        v.DATA_DHW_PUMP_HOURS: 10123,
        v.DATA_DHW_BURNER_HOURS: 99,
    },
    v.OTGW: {},
    v.THERMOSTAT: {
        v.DATA_MASTER_CH_ENABLED: 0,
        v.DATA_MASTER_DHW_ENABLED: 1,
        v.DATA_MASTER_COOLING_ENABLED: 0,
        v.DATA_MASTER_OTC_ENABLED: 1,
        v.DATA_MASTER_CH2_ENABLED: 0,
        v.DATA_CONTROL_SETPOINT: 1.23,
        v.DATA_ROOM_SETPOINT: 5.67,
        v.DATA_COOLING_CONTROL: 2.34,
        v.DATA_CONTROL_SETPOINT_2: 3.45,
        v.DATA_ROOM_SETPOINT_2: 9.09,
        v.DATA_ROOM_TEMP: 0.98,
        v.DATA_VH_MASTER_VENT_ENABLED: 0,
        v.DATA_VH_MASTER_BYPASS_POS: 1,
        v.DATA_VH_MASTER_BYPASS_MODE: 0,
        v.DATA_VH_MASTER_FREE_VENT_MODE: 1,
        v.DATA_VH_CONTROL_SETPOINT: 9,
    },
}

_status_4 = (
    "10101010/01010101,1.23,01010101/10101010,2.34,0/1,3.45,4.56,5.67,6.78,7.89,8.90,"
    "9.09,0.98,12/34,56/78,9.87,8.76,1234,2345,3456,4567,5678,6789,7890,8909"
)
_status_expect_4 = {
    v.BOILER: {
        v.DATA_SLAVE_FAULT_IND: 1,
        v.DATA_SLAVE_CH_ACTIVE: 0,
        v.DATA_SLAVE_DHW_ACTIVE: 1,
        v.DATA_SLAVE_FLAME_ON: 0,
        v.DATA_SLAVE_COOLING_ACTIVE: 1,
        v.DATA_SLAVE_CH2_ACTIVE: 0,
        v.DATA_SLAVE_DIAG_IND: 1,
        v.DATA_REMOTE_TRANSFER_DHW: 1,
        v.DATA_REMOTE_TRANSFER_MAX_CH: 0,
        v.DATA_REMOTE_RW_DHW: 0,
        v.DATA_REMOTE_RW_MAX_CH: 1,
        v.DATA_SLAVE_MAX_RELATIVE_MOD: 2.34,
        v.DATA_SLAVE_MAX_CAPACITY: 0,
        v.DATA_SLAVE_MIN_MOD_LEVEL: 1,
        v.DATA_REL_MOD_LEVEL: 4.56,
        v.DATA_CH_WATER_PRESS: 5.67,
        v.DATA_CH_WATER_TEMP: 7.89,
        v.DATA_DHW_TEMP: 8.90,
        v.DATA_OUTSIDE_TEMP: 9.09,
        v.DATA_RETURN_WATER_TEMP: 0.98,
        v.DATA_SLAVE_DHW_MAX_SETP: 12,
        v.DATA_SLAVE_DHW_MIN_SETP: 34,
        v.DATA_SLAVE_CH_MAX_SETP: 56,
        v.DATA_SLAVE_CH_MIN_SETP: 78,
        v.DATA_DHW_SETPOINT: 9.87,
        v.DATA_MAX_CH_SETPOINT: 8.76,
        v.DATA_TOTAL_BURNER_STARTS: 1234,
        v.DATA_CH_PUMP_STARTS: 2345,
        v.DATA_DHW_PUMP_STARTS: 3456,
        v.DATA_DHW_BURNER_STARTS: 4567,
        v.DATA_TOTAL_BURNER_HOURS: 5678,
        v.DATA_CH_PUMP_HOURS: 6789,
        v.DATA_DHW_PUMP_HOURS: 7890,
        v.DATA_DHW_BURNER_HOURS: 8909,
    },
    v.OTGW: {},
    v.THERMOSTAT: {
        v.DATA_MASTER_CH_ENABLED: 0,
        v.DATA_MASTER_DHW_ENABLED: 1,
        v.DATA_MASTER_COOLING_ENABLED: 0,
        v.DATA_MASTER_OTC_ENABLED: 1,
        v.DATA_MASTER_CH2_ENABLED: 0,
        v.DATA_CONTROL_SETPOINT: 1.23,
        v.DATA_ROOM_SETPOINT: 3.45,
        v.DATA_ROOM_TEMP: 6.78,
    },
}

pygw_status = SimpleNamespace(
    expect_4=_status_expect_4,
    expect_5=_status_expect_5,
    status_4=_status_4,
    status_5=_status_5,
)


pygw_proto_messages = (
    # Invalid message ID
    (
        ("A", 114, None, None, None),
        None,
    ),
    # _get_flag8
    (
        ("T", v.READ_DATA, v.MSG_STATUS, b"\x43", b"\x00"),
        {
            v.BOILER: {},
            v.OTGW: {},
            v.THERMOSTAT: {
                v.DATA_MASTER_CH_ENABLED: 1,
                v.DATA_MASTER_DHW_ENABLED: 1,
                v.DATA_MASTER_COOLING_ENABLED: 0,
                v.DATA_MASTER_OTC_ENABLED: 0,
                v.DATA_MASTER_CH2_ENABLED: 0,
            },
        },
    ),
    # _get_f8_8
    (
        ("B", v.WRITE_ACK, v.MSG_TDHWSET, b"\x14", b"\x80"),
        {v.BOILER: {v.DATA_DHW_SETPOINT: 20.5}, v.OTGW: {}, v.THERMOSTAT: {}},
    ),
    # _get_flag8 with skipped bits
    (
        (
            "R",
            v.READ_ACK,
            v.MSG_STATUSVH,
            b"\00",
            int("01010101", 2).to_bytes(1, "big"),
        ),
        {
            v.BOILER: {
                v.DATA_VH_SLAVE_FAULT_INDICATE: 1,
                v.DATA_VH_SLAVE_VENT_MODE: 0,
                v.DATA_VH_SLAVE_BYPASS_STATUS: 1,
                v.DATA_VH_SLAVE_BYPASS_AUTO_STATUS: 0,
                v.DATA_VH_SLAVE_FREE_VENT_STATUS: 1,
                v.DATA_VH_SLAVE_DIAG_INDICATE: 1,
            },
            v.OTGW: {},
            v.THERMOSTAT: {},
        },
    ),
    # Combined _get_flag8 and _get_u8
    (
        (
            "R",
            v.WRITE_ACK,
            v.MSG_SCONFIG,
            int("10101010", 2).to_bytes(1, "big"),
            b"\xFF",
        ),
        {
            v.BOILER: {
                v.DATA_SLAVE_DHW_PRESENT: 0,
                v.DATA_SLAVE_CONTROL_TYPE: 1,
                v.DATA_SLAVE_COOLING_SUPPORTED: 0,
                v.DATA_SLAVE_DHW_CONFIG: 1,
                v.DATA_SLAVE_MASTER_LOW_OFF_PUMP: 0,
                v.DATA_SLAVE_CH2_PRESENT: 1,
                v.DATA_SLAVE_MEMBERID: 255,
            },
            v.OTGW: {},
            v.THERMOSTAT: {},
        },
    ),
    # _get_u16
    (
        ("A", v.READ_ACK, v.MSG_BURNSTARTS, b"\x12", b"\xAA"),
        {v.BOILER: {}, v.OTGW: {}, v.THERMOSTAT: {v.DATA_TOTAL_BURNER_STARTS: 4778}},
    ),
    # _get_s8
    (
        ("R", v.WRITE_ACK, v.MSG_TCHSETUL, b"\x50", b"\x1E"),
        {
            v.BOILER: {v.DATA_SLAVE_CH_MAX_SETP: 80, v.DATA_SLAVE_CH_MIN_SETP: 30},
            v.OTGW: {},
            v.THERMOSTAT: {},
        },
    ),
    # _get_s16
    (
        ("B", v.READ_ACK, v.MSG_TEXHAUST, b"\xFF", b"\x83"),
        {v.BOILER: {v.DATA_EXHAUST_TEMP: -125}, v.OTGW: {}, v.THERMOSTAT: {}},
    ),
)



"""Global pyotgw values"""

MSG_STATUS = b"\x00"
MSG_TSET = b"\x01"
MSG_MCONFIG = b"\x02"
MSG_SCONFIG = b"\x03"
MSG_COMMAND = b"\x04"
MSG_ASFFLAGS = b"\x05"
MSG_RBPFLAGS = b"\x06"
MSG_COOLING = b"\x07"
MSG_TSETC2 = b"\x08"
MSG_TROVRD = b"\x09"
MSG_TSP = b"\x0A"
MSG_TSPIDX = b"\x0B"
MSG_FHBSIZE = b"\x0C"
MSG_FHBIDX = b"\x0D"
MSG_MAXRMOD = b"\x0E"
MSG_MAXCAPMINMOD = b"\x0F"
MSG_TRSET = b"\x10"
MSG_RELMOD = b"\x11"
MSG_CHPRESS = b"\x12"
MSG_DHWFLOW = b"\x13"
MSG_TIME = b"\x14"
MSG_DATE = b"\x15"
MSG_YEAR = b"\x16"
MSG_TRSET2 = b"\x17"
MSG_TROOM = b"\x18"
MSG_TBOILER = b"\x19"
MSG_TDHW = b"\x1A"
MSG_TOUTSIDE = b"\x1B"
MSG_TRET = b"\x1C"
MSG_TSTOR = b"\x1D"
MSG_TCOLL = b"\x1E"
MSG_TFLOWCH2 = b"\x1F"
MSG_TDHW2 = b"\x20"
MSG_TEXHAUST = b"\x21"
MSG_TDHWSETUL = b"\x30"
MSG_TCHSETUL = b"\x31"
MSG_OTCCURVEUL = b"\x32"
MSG_TDHWSET = b"\x38"
MSG_MAXTSET = b"\x39"
MSG_OTCCURVE = b"\x3A"
MSG_STATUSVH = b"\x46"
MSG_RELVENTPOS = b"\x47"
MSG_RELVENT = b"\x4D"
MSG_ROVRD = b"\x64"
MSG_OEMDIAG = b"\x73"
MSG_BURNSTARTS = b"\x74"
MSG_CHPUMPSTARTS = b"\x75"
MSG_DHWPUMPSTARTS = b"\x76"
MSG_DHWBURNSTARTS = b"\x77"
MSG_BURNHRS = b"\x78"
MSG_CHPUMPHRS = b"\x79"
MSG_DHWPUMPHRS = b"\x7A"
MSG_DHWBURNHRS = b"\x7B"
MSG_OTVERM = b"\x7C"
MSG_OTVERS = b"\x7D"
MSG_MVER = b"\x7E"
MSG_SVER = b"\x7F"

BOILER = "boiler"
OTGW = "gateway"
THERMOSTAT = "thermostat"

DEFAULT_STATUS = {BOILER: {}, OTGW: {}, THERMOSTAT: {}}

# MSG_STATUS
DATA_MASTER_CH_ENABLED = "master_ch_enabled"
DATA_MASTER_DHW_ENABLED = "master_dhw_enabled"
DATA_MASTER_COOLING_ENABLED = "master_cooling_enabled"
DATA_MASTER_OTC_ENABLED = "master_otc_enabled"
DATA_MASTER_CH2_ENABLED = "master_ch2_enabled"
DATA_SLAVE_FAULT_IND = "slave_fault_indication"
DATA_SLAVE_CH_ACTIVE = "slave_ch_active"
DATA_SLAVE_DHW_ACTIVE = "slave_dhw_active"
DATA_SLAVE_FLAME_ON = "slave_flame_on"
DATA_SLAVE_COOLING_ACTIVE = "slave_cooling_active"
DATA_SLAVE_CH2_ACTIVE = "slave_ch2_active"
DATA_SLAVE_DIAG_IND = "slave_diagnostic_indication"

# MSG_TSET
DATA_CONTROL_SETPOINT = "control_setpoint"

# MSG_MCONFIG
DATA_MASTER_MEMBERID = "master_memberid"

# MSG_SCONFIG
DATA_SLAVE_DHW_PRESENT = "slave_dhw_present"
DATA_SLAVE_CONTROL_TYPE = "slave_control_type"
DATA_SLAVE_COOLING_SUPPORTED = "slave_cooling_supported"
DATA_SLAVE_DHW_CONFIG = "slave_dhw_config"
DATA_SLAVE_MASTER_LOW_OFF_PUMP = "slave_master_low_off_pump"
DATA_SLAVE_CH2_PRESENT = "slave_ch2_present"
DATA_SLAVE_MEMBERID = "slave_memberid"

# MSG_COMMAND

# MSG_ASFFLAGS
DATA_SLAVE_SERVICE_REQ = "slave_service_required"
DATA_SLAVE_REMOTE_RESET = "slave_remote_reset"
DATA_SLAVE_LOW_WATER_PRESS = "slave_low_water_pressure"
DATA_SLAVE_GAS_FAULT = "slave_gas_fault"
DATA_SLAVE_AIR_PRESS_FAULT = "slave_air_pressure_fault"
DATA_SLAVE_WATER_OVERTEMP = "slave_water_overtemp"
DATA_SLAVE_OEM_FAULT = "slave_oem_fault"

# MSG_RBPFLAGS
DATA_REMOTE_TRANSFER_DHW = "remote_transfer_dhw"
DATA_REMOTE_TRANSFER_MAX_CH = "remote_transfer_max_ch"
DATA_REMOTE_RW_DHW = "remote_rw_dhw"
DATA_REMOTE_RW_MAX_CH = "remote_rw_max_ch"

# MSG_COOLING
DATA_COOLING_CONTROL = "cooling_control"

# MSG_TSETC2
DATA_CONTROL_SETPOINT_2 = "control_setpoint_2"

# MSG_TROVRD
DATA_ROOM_SETPOINT_OVRD = "room_setpoint_ovrd"

# MSG_TSP

# MSG_TSPIDX

# MSG_FHBSIZE

# MSG_FHBIDX

# MSG_MAXRMOD
DATA_SLAVE_MAX_RELATIVE_MOD = "slave_max_relative_modulation"

# MSG_MAXCAPMINMOD
DATA_SLAVE_MAX_CAPACITY = "slave_max_capacity"
DATA_SLAVE_MIN_MOD_LEVEL = "slave_min_mod_level"

# MSG_TRSET
DATA_ROOM_SETPOINT = "room_setpoint"

# MSG_RELMOD
DATA_REL_MOD_LEVEL = "relative_mod_level"

# MSG_CHPRESS
DATA_CH_WATER_PRESS = "ch_water_pressure"

# MSG_DHWFLOW
DATA_DHW_FLOW_RATE = "dhw_flow_rate"

# MSG_TIME

# MSG_DATE

# MSG_YEAR

# MSG_TRSET2
DATA_ROOM_SETPOINT_2 = "room_setpoint_2"

# MSG_TROOM
DATA_ROOM_TEMP = "room_temp"

# MSG_TBOILER
DATA_CH_WATER_TEMP = "ch_water_temp"

# MSG_TDHW
DATA_DHW_TEMP = "dhw_temp"

# MSG_TOUTSIDE
DATA_OUTSIDE_TEMP = "outside_temp"

# MSG_TRET
DATA_RETURN_WATER_TEMP = "return_water_temp"

# MSG_TSTOR
DATA_SOLAR_STORAGE_TEMP = "solar_storage_temp"

# MSG_TCOLL
DATA_SOLAR_COLL_TEMP = "solar_coll_temp"

# MSG_TFLOWCH2
DATA_CH_WATER_TEMP_2 = "ch_water_temp_2"

# MSG_TDHW2
DATA_DHW_TEMP_2 = "dhw_temp_2"

# MSG_TEXHAUST
DATA_EXHAUST_TEMP = "exhaust_temp"

# MSG_TDHWSETUL
DATA_SLAVE_DHW_MAX_SETP = "slave_dhw_max_setp"
DATA_SLAVE_DHW_MIN_SETP = "slave_dhw_min_setp"

# MSG_TCHSETUL
DATA_SLAVE_CH_MAX_SETP = "slave_ch_max_setp"
DATA_SLAVE_CH_MIN_SETP = "slave_ch_min_setp"

# MSG_OTCCURVEUL

# MSG_TDHWSET
DATA_DHW_SETPOINT = "dhw_setpoint"

# MSG_MAXTSET
DATA_MAX_CH_SETPOINT = "max_ch_setpoint"

# MSG_OTCCURVE

# MSG_STATUSVH
DATA_VH_MASTER_VENT_ENABLED = "vh_master_vent_enabled"
DATA_VH_MASTER_BYPASS_POS = "vh_master_bypass_pos"
DATA_VH_MASTER_BYPASS_MODE = "vh_master_bypass_mode"
DATA_VH_MASTER_FREE_VENT_MODE = "vh_master_free_vent_mode"
DATA_VH_SLAVE_FAULT_INDICATE = "vh_slave_fault_indicate"
DATA_VH_SLAVE_VENT_MODE = "vh_slave_vent_mode"
DATA_VH_SLAVE_BYPASS_STATUS = "vh_slave_bypass_status"
DATA_VH_SLAVE_BYPASS_AUTO_STATUS = "vh_slave_bypass_auto_status"
DATA_VH_SLAVE_FREE_VENT_STATUS = "vh_slave_free_vent_status"
DATA_VH_SLAVE_DIAG_INDICATE = "vh_slave_diag_indicate"

# MSG_RELVENTPOS
DATA_VH_CONTROL_SETPOINT = "vh_control_setpoint"

# MSG_RELVENT
DATA_VH_RELATIVE_VENT = "vh_relative_vent"

# MSG_ROVRD
DATA_ROVRD_MAN_PRIO = "rovrd_man_prio"
DATA_ROVRD_AUTO_PRIO = "rovrd_auto_prio"

# MSG_OEMDIAG
DATA_OEM_DIAG = "oem_diag"

# MSG_BURNSTARTS
DATA_TOTAL_BURNER_STARTS = "burner_starts"

# MSG_CHPUMPSTARTS
DATA_CH_PUMP_STARTS = "ch_pump_starts"

# MSG_DHWPUMPSTARTS
DATA_DHW_PUMP_STARTS = "dhw_pump_starts"

# MSG_DHWBURNSTARTS
DATA_DHW_BURNER_STARTS = "dhw_burner_starts"

# MSG_BURNHRS
DATA_TOTAL_BURNER_HOURS = "burner_hours"

# MSG_CHPUMPHRS
DATA_CH_PUMP_HOURS = "ch_pump_hours"

# MSG_DHWPUMPHRS
DATA_DHW_PUMP_HOURS = "dhw_pump_hours"

# MSG_DHWBURNHRS
DATA_DHW_BURNER_HOURS = "dhw_burner_hours"

# MSG_OTVERM
DATA_MASTER_OT_VERSION = "master_ot_version"

# MSG_OTVERS
DATA_SLAVE_OT_VERSION = "slave_ot_version"

# MSG_MVER
DATA_MASTER_PRODUCT_TYPE = "master_product_type"
DATA_MASTER_PRODUCT_VERSION = "master_product_version"

# MSG_SVER
DATA_SLAVE_PRODUCT_TYPE = "slave_product_type"
DATA_SLAVE_PRODUCT_VERSION = "slave_product_version"


READ_DATA = 0x0
WRITE_DATA = 0x1
INVALID_DATA = 0x2
RESERVED = 0x3
READ_ACK = 0x4
WRITE_ACK = 0x5
DATA_INVALID = 0x6
UNKNOWN_DATAID = 0x7

OTGW_DEFAULT_TIMEOUT = 3

OTGW_CMD_TARGET_TEMP = "TT"
OTGW_CMD_TARGET_TEMP_CONST = "TC"
OTGW_CMD_OUTSIDE_TEMP = "OT"
OTGW_CMD_SET_CLOCK = "SC"
OTGW_CMD_HOT_WATER = "HW"
OTGW_CMD_REPORT = "PR"
OTGW_CMD_SUMMARY = "PS"
OTGW_CMD_MODE = "GW"
OTGW_CMD_LED_A = "LA"
OTGW_CMD_LED_B = "LB"
OTGW_CMD_LED_C = "LC"
OTGW_CMD_LED_D = "LD"
OTGW_CMD_LED_E = "LE"
OTGW_CMD_LED_F = "LF"
OTGW_CMD_GPIO_A = "GA"
OTGW_CMD_GPIO_B = "GB"
OTGW_CMD_SETBACK = "SB"
OTGW_CMD_TEMP_SENSOR = "TS"
OTGW_CMD_ADD_ALT = "AA"
OTGW_CMD_DEL_ALT = "DA"
OTGW_CMD_UNKNOWN_ID = "UI"
OTGW_CMD_KNOWN_ID = "KI"
OTGW_CMD_PRIO_MSG = "PM"
OTGW_CMD_SET_RESP = "SR"
OTGW_CMD_CLR_RESP = "CR"
OTGW_CMD_SET_MAX = "SH"
OTGW_CMD_SET_WATER = "SW"
OTGW_CMD_MAX_MOD = "MM"
OTGW_CMD_CONTROL_SETPOINT = "CS"
OTGW_CMD_CONTROL_SETPOINT_2 = "C2"
OTGW_CMD_CONTROL_HEATING = "CH"
OTGW_CMD_CONTROL_HEATING_2 = "H2"
OTGW_CMD_VENT = "VS"
OTGW_CMD_RST_CNT = "RS"
OTGW_CMD_IGNORE_TRANS = "IT"
OTGW_CMD_OVRD_HIGH = "OH"
OTGW_CMD_OVRD_THRMST = "FT"
OTGW_CMD_VREF = "VR"

OTGW_MODE = "otgw_mode"
OTGW_DHW_OVRD = "otgw_dhw_ovrd"
OTGW_ABOUT = "otgw_about"
OTGW_BUILD = "otgw_build"
OTGW_CLOCKMHZ = "otgw_clockmhz"
OTGW_LED_A = "otgw_led_a"
OTGW_LED_B = "otgw_led_b"
OTGW_LED_C = "otgw_led_c"
OTGW_LED_D = "otgw_led_d"
OTGW_LED_E = "otgw_led_e"
OTGW_LED_F = "otgw_led_f"
OTGW_GPIO_A = "otgw_gpio_a"
OTGW_GPIO_B = "otgw_gpio_b"
OTGW_GPIO_A_STATE = "otgw_gpio_a_state"
OTGW_GPIO_B_STATE = "otgw_gpio_b_state"
OTGW_RST_CAUSE = "otgw_reset_cause"
OTGW_SB_TEMP = "otgw_setback_temp"
OTGW_SETP_OVRD_MODE = "otgw_setpoint_ovrd_mode"
OTGW_SMART_PWR = "otgw_smart_pwr"
OTGW_TEMP_SENSOR = "otgw_temp_sensor"
OTGW_THRM_DETECT = "otgw_thermostat_detect"
OTGW_IGNORE_TRANSITIONS = "otgw_ignore_transitions"
OTGW_OVRD_HB = "otgw_ovrd_high_byte"
OTGW_VREF = "otgw_vref"

OTGW_SETP_OVRD_TEMPORARY = "T"
OTGW_SETP_OVRD_PERMANENT = "C"
OTGW_SETP_OVRD_DISABLED = "N"
OTGW_MODE_MONITOR = "M"
OTGW_MODE_GATEWAY = "G"
OTGW_MODE_RESET = "R"

OTGW_REPORT_ABOUT = "A"
OTGW_REPORT_BUILDDATE = "B"
OTGW_REPORT_CLOCKMHZ = "C"
OTGW_REPORT_TEMP_SENSOR = "D"
OTGW_REPORT_GPIO_FUNCS = "G"
OTGW_REPORT_GPIO_STATES = "I"
OTGW_REPORT_LED_FUNCS = "L"
OTGW_REPORT_GW_MODE = "M"
OTGW_REPORT_SETPOINT_OVRD = "O"
OTGW_REPORT_SMART_PWR = "P"
OTGW_REPORT_RST_CAUSE = "Q"
OTGW_REPORT_THERMOSTAT_DETECT = "R"
OTGW_REPORT_SETBACK_TEMP = "S"
OTGW_REPORT_TWEAKS = "T"
OTGW_REPORT_VREF = "V"
OTGW_REPORT_DHW_SETTING = "W"

OTGW_REPORTS = {
    OTGW_REPORT_ABOUT: OTGW_ABOUT,
    OTGW_REPORT_BUILDDATE: OTGW_BUILD,
    OTGW_REPORT_CLOCKMHZ: OTGW_CLOCKMHZ,
    OTGW_REPORT_DHW_SETTING: OTGW_DHW_OVRD,
    OTGW_REPORT_GPIO_FUNCS: [OTGW_GPIO_A, OTGW_GPIO_B],
    OTGW_REPORT_GPIO_STATES: [OTGW_GPIO_A_STATE, OTGW_GPIO_B_STATE],
    OTGW_REPORT_LED_FUNCS: [
        OTGW_LED_A,
        OTGW_LED_B,
        OTGW_LED_C,
        OTGW_LED_D,
        OTGW_LED_E,
        OTGW_LED_F,
    ],
    OTGW_REPORT_GW_MODE: OTGW_MODE,
    OTGW_REPORT_RST_CAUSE: OTGW_RST_CAUSE,
    OTGW_REPORT_SETBACK_TEMP: OTGW_SB_TEMP,
    OTGW_REPORT_SETPOINT_OVRD: OTGW_SETP_OVRD_MODE,
    OTGW_REPORT_SMART_PWR: OTGW_SMART_PWR,
    OTGW_REPORT_TEMP_SENSOR: OTGW_TEMP_SENSOR,
    OTGW_REPORT_THERMOSTAT_DETECT: OTGW_THRM_DETECT,
    OTGW_REPORT_TWEAKS: [OTGW_IGNORE_TRANSITIONS, OTGW_OVRD_HB],
    OTGW_REPORT_VREF: OTGW_VREF,
}

NO_GOOD = "NG"
SYNTAX_ERR = "SE"
BAD_VALUE = "BV"
OUT_OF_RANGE = "OR"
NO_SPACE = "NS"
NOT_FOUND = "NF"
OVERRUN_ERR = "OE"
MPC_ERR = "MPC"

OTGW_ERRS = {
    NO_GOOD: RuntimeError(
        "No Good: The command code is unknown or unsupported on this "
        "version of the OpenTherm Gateway."
    ),
    SYNTAX_ERR: SyntaxError(
        "Syntax Error: The command contained an "
        "unexpected character or was incomplete."
    ),
    BAD_VALUE: ValueError(
        "Bad Value: The command contained a data value that is not allowed or not "
        "supported on this version of the OpenTherm Gateway."
    ),
    OUT_OF_RANGE: RuntimeError(
        "Out of Range: A number was specified outside of the allowed range."
    ),
    NO_SPACE: RuntimeError(
        "No Space: The alternative Data-ID could not be "
        "added because the table is full."
    ),
    NOT_FOUND: RuntimeError(
        "Not Found: The specified alternative Data-ID "
        "could not be removed because it does not exist "
        "in the table."
    ),
    OVERRUN_ERR: RuntimeError(
        "Overrun Error: The processor was busy and "
        "failed to process all received characters."
    ),
    MPC_ERR: RuntimeError("MPC Error"),
}



"""Tests for pyotgw/messages.py"""

import pyotgw.messages as m
from pyotgw.messageprocessor import MessageProcessor


def test_message_registry():
    """Test message registry values."""
    for msgid, processing in m.REGISTRY.items():
        assert 0 <= int.from_bytes(msgid, "big") < 128
        assert isinstance(processing[m.M2S], list)
        assert isinstance(processing[m.S2M], list)

        for action in [*processing[m.M2S], *processing[m.S2M]]:
            assert hasattr(MessageProcessor, action[m.FUNC])
            assert isinstance(action[m.ARGS], tuple)
            assert isinstance(action[m.RETURNS], tuple)



import os

from setuptools import setup


# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="pyotgw",
    version="2.2.2",
    author="Milan van Nugteren",
    author_email="milan@network23.nl",
    description=(
        "A library to interface with the opentherm gateway through "
        "serial or network connection."
    ),
    license="GPLv3+",
    keywords="opentherm gateway otgw",
    url="https://github.com/mvn23/pyotgw",
    packages=["pyotgw"],
    python_requires=">=3.8",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    install_requires=["pyserial-asyncio-fast"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Topic :: Software Development :: Libraries",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
    ],
)



"""Tests for pyotgw/commandprocessor.py"""

import asyncio
import logging
from unittest.mock import MagicMock, call, patch

import pytest

import pyotgw.vars as v
from tests.helpers import called_once, let_queue_drain


@pytest.mark.asyncio
async def test_submit_response_queuefull(caplog, pygw_proto):
    """Test queuefull on submit_response()"""
    test_lines = ("BCDEF", "A1A2B3C4D", "MustBeCommand", "AlsoCommand")
    with patch.object(
        pygw_proto.command_processor._cmdq, "put_nowait", side_effect=asyncio.QueueFull
    ) as put_nowait, caplog.at_level(logging.ERROR):
        pygw_proto.line_received(test_lines[3])

    pygw_proto.activity_callback.assert_called_once()
    put_nowait.assert_called_once_with(test_lines[3])
    assert pygw_proto.command_processor._cmdq.qsize() == 0
    assert caplog.record_tuples == [
        (
            "pyotgw.commandprocessor",
            logging.ERROR,
            f"Queue full, discarded message: {test_lines[3]}",
        ),
    ]


@pytest.mark.asyncio
async def test_issue_cmd(caplog, pygw_proto):
    """Test OpenThermProtocol.issue_cmd()"""
    pygw_proto._connected = False
    with caplog.at_level(logging.DEBUG):
        assert await pygw_proto.command_processor.issue_cmd("PS", 1, 0) is None

    assert caplog.record_tuples == [
        (
            "pyotgw.commandprocessor",
            logging.DEBUG,
            "Serial transport closed, not sending command PS",
        ),
    ]
    caplog.clear()

    loop = asyncio.get_running_loop()
    pygw_proto._connected = True
    pygw_proto.command_processor._cmdq.put_nowait("thisshouldbecleared")
    pygw_proto.transport.write = MagicMock()

    with caplog.at_level(logging.DEBUG):
        task = loop.create_task(
            pygw_proto.command_processor.issue_cmd(
                v.OTGW_CMD_REPORT,
                "I",
                1,
            )
        )
        await let_queue_drain(pygw_proto.command_processor._cmdq)

        pygw_proto.transport.write.assert_called_once_with(b"PR=I\r\n")
        assert caplog.record_tuples == [
            (
                "pyotgw.commandprocessor",
                logging.DEBUG,
                "Clearing leftover message from command queue: thisshouldbecleared",
            ),
            (
                "pyotgw.commandprocessor",
                logging.DEBUG,
                "Sending command: PR with value I",
            ),
        ]
        caplog.clear()

        pygw_proto.command_processor.submit_response("SE")
        pygw_proto.command_processor.submit_response("SE")
        with pytest.raises(SyntaxError):
            await task

    assert pygw_proto.transport.write.call_args_list == [
        call(b"PR=I\r\n"),
        call(b"PR=I\r\n"),
    ]

    assert caplog.record_tuples == [
        (
            "pyotgw.commandprocessor",
            logging.DEBUG,
            "Response submitted. Queue size: 1",
        ),
        (
            "pyotgw.commandprocessor",
            logging.DEBUG,
            "Response submitted. Queue size: 2",
        ),
        (
            "pyotgw.commandprocessor",
            logging.DEBUG,
            "Got possible response for command PR: SE",
        ),
        (
            "pyotgw.commandprocessor",
            logging.WARNING,
            "Command PR failed with SE, retrying...",
        ),
        (
            "pyotgw.commandprocessor",
            logging.DEBUG,
            "Got possible response for command PR: SE",
        ),
    ]
    caplog.clear()

    pygw_proto.transport.write = MagicMock()
    with caplog.at_level(logging.WARNING):
        task = loop.create_task(
            pygw_proto.command_processor.issue_cmd(
                v.OTGW_CMD_CONTROL_SETPOINT_2,
                20.501,
                1,
            )
        )
        await called_once(pygw_proto.transport.write)
        pygw_proto.transport.write.assert_called_once_with(b"C2=20.50\r\n")
        pygw_proto.command_processor.submit_response("InvalidCommand")
        pygw_proto.command_processor.submit_response("C2: 20.50")
        assert await task == "20.50"

    assert pygw_proto.transport.write.call_args_list == [
        call(b"C2=20.50\r\n"),
        call(b"C2=20.50\r\n"),
    ]
    assert caplog.record_tuples == [
        (
            "pyotgw.commandprocessor",
            logging.WARNING,
            "Unknown message in command queue: InvalidCommand",
        ),
        (
            "pyotgw.commandprocessor",
            logging.WARNING,
            "Command C2 failed with InvalidCommand, retrying...",
        ),
    ]
    caplog.clear()

    pygw_proto.transport.write = MagicMock()
    with caplog.at_level(logging.WARNING):
        task = loop.create_task(
            pygw_proto.command_processor.issue_cmd(
                v.OTGW_CMD_CONTROL_HEATING_2,
                -1,
                2,
            )
        )
        await called_once(pygw_proto.transport.write)
        pygw_proto.transport.write.assert_called_once_with(b"H2=-1\r\n")
        pygw_proto.command_processor.submit_response("Error 03")
        pygw_proto.command_processor.submit_response("H2: BV")
        pygw_proto.command_processor.submit_response("H2: BV")
        with pytest.raises(ValueError):
            await task

    assert caplog.record_tuples == [
        (
            "pyotgw.commandprocessor",
            logging.WARNING,
            "Received Error 03. "
            "If this happens during a reset of the gateway it can be safely ignored.",
        ),
        (
            "pyotgw.commandprocessor",
            logging.WARNING,
            "Command H2 failed with Error 03, retrying...",
        ),
        (
            "pyotgw.commandprocessor",
            logging.WARNING,
            "Command H2 failed with H2: BV, retrying...",
        ),
    ]

    pygw_proto.transport.write = MagicMock()
    task = loop.create_task(
        pygw_proto.command_processor.issue_cmd(
            v.OTGW_CMD_MODE,
            "R",
            0,
        )
    )
    await called_once(pygw_proto.transport.write)
    pygw_proto.command_processor.submit_response("ThisGetsIgnored")
    pygw_proto.command_processor.submit_response("OpenTherm Gateway 4.3.5")

    assert await task is True

    pygw_proto.transport.write = MagicMock()
    task = loop.create_task(
        pygw_proto.command_processor.issue_cmd(
            v.OTGW_CMD_SUMMARY,
            1,
            0,
        )
    )
    await called_once(pygw_proto.transport.write)
    pygw_proto.command_processor.submit_response("PS: 1")
    pygw_proto.command_processor.submit_response(
        "part_2_will_normally_be_parsed_by_get_status",
    )

    assert await task == ["1", "part_2_will_normally_be_parsed_by_get_status"]



"""Tests for pyotgw/status.py"""
import asyncio
import logging
from unittest.mock import MagicMock

import pytest

import pyotgw.vars as v
from tests.helpers import called_once


def test_reset(pygw_status):
    """Test StatusManager.reset()"""
    assert pygw_status.status == v.DEFAULT_STATUS

    pygw_status.submit_partial_update(v.OTGW, {"Test": "value"})

    assert pygw_status.status != v.DEFAULT_STATUS
    assert not pygw_status._updateq.empty()

    pygw_status.reset()

    assert pygw_status.status == v.DEFAULT_STATUS
    assert pygw_status._updateq.empty()


def test_status(pygw_status):
    """Test StatusManager.status()"""
    assert pygw_status.status == pygw_status._status
    assert pygw_status.status is not pygw_status._status


def test_delete_value(pygw_status):
    """Test StatusManager.delete_value()"""
    assert not pygw_status.delete_value("Invalid", v.OTGW_MODE)
    assert not pygw_status.delete_value(v.OTGW, v.OTGW_MODE)

    pygw_status.submit_partial_update(v.THERMOSTAT, {v.DATA_ROOM_SETPOINT: 20.5})
    pygw_status._updateq.get_nowait()

    assert pygw_status.delete_value(v.THERMOSTAT, v.DATA_ROOM_SETPOINT)
    assert pygw_status._updateq.get_nowait() == v.DEFAULT_STATUS


def test_submit_partial_update(caplog, pygw_status):
    """Test StatusManager.submit_partial_update()"""
    with caplog.at_level(logging.ERROR):
        assert not pygw_status.submit_partial_update("Invalid", {})

    assert pygw_status._updateq.empty()
    assert caplog.record_tuples == [
        (
            "pyotgw.status",
            logging.ERROR,
            "Invalid status part for update: Invalid",
        ),
    ]
    caplog.clear()

    with caplog.at_level(logging.ERROR):
        assert not pygw_status.submit_partial_update(v.OTGW, "Invalid")

    assert pygw_status._updateq.empty()
    assert caplog.record_tuples == [
        (
            "pyotgw.status",
            logging.ERROR,
            f"Update for {v.OTGW} is not a dict: Invalid",
        ),
    ]
    caplog.clear()

    pygw_status.submit_partial_update(v.BOILER, {v.DATA_CONTROL_SETPOINT: 1.5})
    pygw_status.submit_partial_update(v.OTGW, {v.OTGW_ABOUT: "test value"})
    pygw_status.submit_partial_update(v.THERMOSTAT, {v.DATA_ROOM_SETPOINT: 20})

    assert pygw_status.status == {
        v.BOILER: {v.DATA_CONTROL_SETPOINT: 1.5},
        v.OTGW: {v.OTGW_ABOUT: "test value"},
        v.THERMOSTAT: {v.DATA_ROOM_SETPOINT: 20},
    }
    assert pygw_status._updateq.qsize() == 3
    assert pygw_status._updateq.get_nowait() == {
        v.BOILER: {v.DATA_CONTROL_SETPOINT: 1.5},
        v.OTGW: {},
        v.THERMOSTAT: {},
    }
    assert pygw_status._updateq.get_nowait() == {
        v.BOILER: {v.DATA_CONTROL_SETPOINT: 1.5},
        v.OTGW: {v.OTGW_ABOUT: "test value"},
        v.THERMOSTAT: {},
    }
    assert pygw_status._updateq.get_nowait() == {
        v.BOILER: {v.DATA_CONTROL_SETPOINT: 1.5},
        v.OTGW: {v.OTGW_ABOUT: "test value"},
        v.THERMOSTAT: {v.DATA_ROOM_SETPOINT: 20},
    }


def test_submit_full_update(caplog, pygw_status):
    """Test StatusManager.submit_full_update()"""
    assert pygw_status.submit_full_update({})
    assert pygw_status._updateq.qsize() == 1
    assert pygw_status._updateq.get_nowait() == v.DEFAULT_STATUS

    with caplog.at_level(logging.ERROR):
        pygw_status.submit_full_update({"Invalid": {}})

    assert pygw_status._updateq.empty()
    assert caplog.record_tuples == [
        (
            "pyotgw.status",
            logging.ERROR,
            "Invalid status part for update: Invalid",
        ),
    ]
    caplog.clear()

    with caplog.at_level(logging.ERROR):
        pygw_status.submit_full_update({v.OTGW: "Invalid"})

    assert pygw_status._updateq.empty()
    assert caplog.record_tuples == [
        (
            "pyotgw.status",
            logging.ERROR,
            f"Update for {v.OTGW} is not a dict: Invalid",
        ),
    ]
    caplog.clear()

    pygw_status.submit_full_update(
        {
            v.BOILER: {v.DATA_CONTROL_SETPOINT: 1.5},
            v.OTGW: {v.OTGW_ABOUT: "test value"},
            v.THERMOSTAT: {v.DATA_ROOM_SETPOINT: 20},
        }
    )

    assert pygw_status.status == {
        v.BOILER: {v.DATA_CONTROL_SETPOINT: 1.5},
        v.OTGW: {v.OTGW_ABOUT: "test value"},
        v.THERMOSTAT: {v.DATA_ROOM_SETPOINT: 20},
    }
    assert pygw_status._updateq.qsize() == 1
    assert pygw_status._updateq.get_nowait() == {
        v.BOILER: {v.DATA_CONTROL_SETPOINT: 1.5},
        v.OTGW: {v.OTGW_ABOUT: "test value"},
        v.THERMOSTAT: {v.DATA_ROOM_SETPOINT: 20},
    }


def test_subscribe(pygw_status):
    """Test StatusManager.subscribe()"""

    def empty_callback():
        return

    assert pygw_status.subscribe(empty_callback)
    assert empty_callback in pygw_status._notify
    assert not pygw_status.subscribe(empty_callback)


def test_unsubscribe(pygw_status):
    """Test StatusManager.unsubscribe()"""

    def empty_callback():
        return

    assert not pygw_status.unsubscribe(empty_callback)
    pygw_status.subscribe(empty_callback)
    assert pygw_status.unsubscribe(empty_callback)
    assert empty_callback not in pygw_status._notify


@pytest.mark.asyncio
async def test_stop_reporting(pygw_status):
    """Test StatusManager.stop_reporting()"""
    assert isinstance(pygw_status._update_task, asyncio.Task)
    await pygw_status.cleanup()
    assert pygw_status._update_task is None


@pytest.mark.asyncio
async def test_process_updates(caplog, pygw_status):
    """Test StatusManager._process_updates()"""

    await pygw_status.cleanup()
    pygw_status.__init__()
    with caplog.at_level(logging.DEBUG):
        # Let the reporting routine start
        await asyncio.sleep(0)

    assert isinstance(pygw_status._update_task, asyncio.Task)
    assert caplog.record_tuples == [
        (
            "pyotgw.status",
            logging.DEBUG,
            "Starting reporting routine",
        ),
    ]
    caplog.clear()

    async def empty_callback_1(status):
        return

    async def empty_callback_2(status):
        return

    mock_callback_1 = MagicMock(side_effect=empty_callback_1)
    mock_callback_2 = MagicMock(side_effect=empty_callback_2)

    pygw_status.subscribe(mock_callback_1)
    pygw_status.subscribe(mock_callback_2)
    pygw_status.submit_partial_update(v.OTGW, {v.OTGW_ABOUT: "Test Value"})
    await asyncio.gather(called_once(mock_callback_1), called_once(mock_callback_2))

    for mock in (mock_callback_1, mock_callback_2):
        mock.assert_called_once_with(
            {
                v.BOILER: {},
                v.OTGW: {v.OTGW_ABOUT: "Test Value"},
                v.THERMOSTAT: {},
            }
        )



"""Tests for pyotgw/protocol.py"""

import asyncio
import logging
from unittest.mock import patch

import pytest

import pyotgw.vars as v
from tests.helpers import called_x_times


def test_connection_made(pygw_proto):
    """Test OpenThermProtocol.connection_made()"""
    # pygw_proto already calls connection_made()
    assert pygw_proto.connected


def test_connection_lost(caplog, pygw_proto):
    """Test OpenThermProtocol.connection_lost()"""
    pygw_proto._received_lines = 1
    pygw_proto.command_processor.submit_response("test cmdq")
    assert not pygw_proto.command_processor._cmdq.empty()

    with caplog.at_level(logging.ERROR):
        pygw_proto.connection_lost(None)

    assert not pygw_proto.active
    assert pygw_proto.command_processor._cmdq.empty()
    assert pygw_proto.status_manager.status == v.DEFAULT_STATUS
    assert caplog.record_tuples == [
        (
            "pyotgw.protocol",
            logging.ERROR,
            "Disconnected: None",
        ),
    ]


def test_connected(pygw_proto):
    """Test OpenThermProtocol.connected()"""
    assert pygw_proto.connected is True
    pygw_proto._connected = False
    assert pygw_proto.connected is False


@pytest.mark.asyncio
async def test_cleanup(pygw_proto):
    """Test OpenThermProtocol.cleanup()"""
    with patch.object(
        pygw_proto.message_processor,
        "cleanup",
    ) as message_processor_cleanup:
        await pygw_proto.cleanup()
    message_processor_cleanup.assert_called_once()


def test_disconnect(pygw_proto):
    """Test OpenThermProtocol.disconnect()"""
    pygw_proto.disconnect()

    pygw_proto._connected = True
    pygw_proto.transport.is_closing.return_value = False

    pygw_proto.disconnect()

    assert not pygw_proto.connected
    pygw_proto.transport.close.assert_called_once()


def test_data_received(caplog, pygw_proto):
    """Test OpenThermProtocol.data_received()"""
    test_input = (
        b"ignorethis\x04A123",
        b"45678\r\n",
        b"\x80\r\n",
    )
    with patch.object(pygw_proto, "line_received") as line_received:
        pygw_proto.data_received(test_input[0])

        line_received.assert_not_called()

        pygw_proto.data_received(test_input[1])
        line_received.assert_called_once_with("A12345678")

        with caplog.at_level(logging.DEBUG):
            pygw_proto.data_received(test_input[2])
        assert pygw_proto._readbuf == b""
        assert caplog.record_tuples == [
            (
                "pyotgw.protocol",
                logging.DEBUG,
                "Invalid data received, ignoring...",
            ),
        ]


@pytest.mark.asyncio
async def test_line_received(caplog, pygw_proto):
    """Test OpenThermProtocol.line_received()"""
    test_lines = ("BCDEF", "A1A2B3C4D", "MustBeCommand", "AlsoCommand")

    with caplog.at_level(logging.DEBUG):
        pygw_proto.line_received(test_lines[0])

    pygw_proto.activity_callback.assert_called_once()
    assert not pygw_proto.active
    assert caplog.record_tuples == [
        (
            "pyotgw.protocol",
            logging.DEBUG,
            f"Received line 1: {test_lines[0]}",
        ),
        (
            "pyotgw.protocol",
            logging.DEBUG,
            f"Ignoring line: {test_lines[0]}",
        ),
    ]

    pygw_proto.activity_callback.reset_mock()
    caplog.clear()

    with patch.object(
        pygw_proto.message_processor,
        "submit_matched_message",
    ) as submit_message, caplog.at_level(logging.DEBUG):
        pygw_proto.line_received(test_lines[1])

    assert pygw_proto.active
    pygw_proto.activity_callback.assert_called_once()
    submit_message.assert_called_once()
    assert caplog.record_tuples == [
        (
            "pyotgw.protocol",
            logging.DEBUG,
            f"Received line 1: {test_lines[1]}",
        ),
    ]

    pygw_proto.activity_callback.reset_mock()
    caplog.clear()

    with caplog.at_level(logging.DEBUG):
        pygw_proto.line_received(test_lines[2])

    assert pygw_proto.active
    pygw_proto.activity_callback.assert_called_once()
    assert pygw_proto.command_processor._cmdq.qsize() == 1
    assert pygw_proto.command_processor._cmdq.get_nowait() == test_lines[2]
    assert caplog.record_tuples == [
        (
            "pyotgw.protocol",
            logging.DEBUG,
            f"Received line 2: {test_lines[2]}",
        ),
        (
            "pyotgw.protocol",
            logging.DEBUG,
            "Submitting line 2 to CommandProcessor",
        ),
        (
            "pyotgw.commandprocessor",
            logging.DEBUG,
            "Response submitted. Queue size: 1",
        ),
    ]

    pygw_proto.activity_callback.reset_mock()
    caplog.clear()


def test_active(pygw_proto):
    """Test OpenThermProtocol.active()"""
    pygw_proto._received_lines = 0
    assert pygw_proto.active is False
    pygw_proto._received_lines = 1
    assert pygw_proto.active is True


@pytest.mark.asyncio
async def test_init_and_wait_for_activity(pygw_proto):
    """Test OpenThermProtocol.init_and_wait_for_activity()"""
    loop = asyncio.get_running_loop()

    with patch.object(pygw_proto.command_processor, "issue_cmd") as issue_cmd:
        task = loop.create_task(pygw_proto.init_and_wait_for_activity())
        await called_x_times(issue_cmd, 1)
    pygw_proto._received_lines = 1
    await task
